from __future__ import annotations

import uuid
from collections.abc import AsyncIterator, Iterator
from contextlib import asynccontextmanager, contextmanager
from datetime import datetime
from typing import TYPE_CHECKING, Any

from distributed.compatibility import PeriodicCallback

if TYPE_CHECKING:
    # Optional runtime dependencies
    import pandas as pd

    # Circular dependencies
    from distributed.client import Client
    from distributed.scheduler import Scheduler


class MemorySampler:
    """Sample cluster-wide memory usage every <interval> seconds.

    **Usage**

    .. code-block:: python

       client = Client()
       ms = MemorySampler()

       with ms.sample("run 1"):
           <run first workflow>
       with ms.sample("run 2"):
           <run second workflow>
       ...
       ms.plot()

    or with an asynchronous client:

    .. code-block:: python

       client = await Client(asynchronous=True)
       ms = MemorySampler()

       async with ms.sample("run 1"):
           <run first workflow>
       async with ms.sample("run 2"):
           <run second workflow>
       ...
       ms.plot()
    """

    samples: dict[str, list[tuple[float, int]]]

    def __init__(self):
        self.samples = {}

    def sample(
        self,
        label: str | None = None,
        *,
        client: Client | None = None,
        measure: str = "process",
        interval: float = 0.5,
    ) -> Any:
        """Context manager that records memory usage in the cluster.
        This is synchronous if the client is synchronous and
        asynchronous if the client is asynchronous.

        The samples are recorded in ``self.samples[<label>]``.

        Parameters
        ==========
        label: str, optional
            Tag to record the samples under in the self.samples dict.
            Default: automatically generate a random label
        client: Client, optional
            client used to connect to the scheduler.
            Default: use the global client
        measure: str, optional
            One of the measures from :class:`distributed.scheduler.MemoryState`.
            Default: sample process memory
        interval: float, optional
            sampling interval, in seconds.
            Default: 0.5
        """
        if not client:
            from distributed.client import get_client

            client = get_client()

        if client.asynchronous:
            return self._sample_async(label, client, measure, interval)
        else:
            return self._sample_sync(label, client, measure, interval)

    @contextmanager
    def _sample_sync(
        self, label: str | None, client: Client, measure: str, interval: float
    ) -> Iterator[None]:
        key = client.sync(
            client.scheduler.memory_sampler_start,
            client=client.id,
            measure=measure,
            interval=interval,
        )
        try:
            yield
        finally:
            samples = client.sync(client.scheduler.memory_sampler_stop, key=key)
            self.samples[label or key] = samples

    @asynccontextmanager
    async def _sample_async(
        self, label: str | None, client: Client, measure: str, interval: float
    ) -> AsyncIterator[None]:
        key = await client.scheduler.memory_sampler_start(
            client=client.id, measure=measure, interval=interval
        )
        try:
            yield
        finally:
            samples = await client.scheduler.memory_sampler_stop(key=key)
            self.samples[label or key] = samples

    def to_pandas(self, *, align: bool = False) -> pd.DataFrame:
        """Return the data series as a pandas.Dataframe.

        Parameters
        ==========
        align : bool, optional
            If True, change the absolute timestamps into time deltas from the first
            sample of each series, so that different series can be visualized side by
            side. If False (the default), use absolute timestamps.
        """
        import pandas as pd

        ss = {}
        for label, s_list in self.samples.items():
            assert s_list  # There's always at least one sample
            s = pd.DataFrame(s_list).set_index(0)[1]
            s.index = pd.to_datetime(s.index, unit="s")
            s.name = label
            if align:
                # convert datetime to timedelta from the first sample
                s.index -= s.index[0]
            ss[label] = s

        df = pd.DataFrame(ss)

        if len(ss) > 1:
            # Forward-fill NaNs in the middle of a series created either by overlapping
            # sampling time range or by align=True. Do not ffill series beyond their
            # last sample.
            df = df.ffill().where(~pd.isna(df.bfill()))

        return df

    def plot(self, *, align: bool = False, **kwargs: Any) -> Any:
        """Plot data series collected so far

        Parameters
        ==========
        align : bool (optional)
            See :meth:`~distributed.diagnostics.MemorySampler.to_pandas`
        kwargs
            Passed verbatim to :meth:`pandas.DataFrame.plot`

        Returns
        =======
        Output of :meth:`pandas.DataFrame.plot`
        """
        df = self.to_pandas(align=align) / 2**30
        return df.plot(
            xlabel="time",
            ylabel="Cluster memory (GiB)",
            **kwargs,
        )


class MemorySamplerExtension:
    """Scheduler extension - server side of MemorySampler"""

    scheduler: Scheduler
    samples: dict[str, list[tuple[float, int]]]

    def __init__(self, scheduler: Scheduler):
        self.scheduler = scheduler
        self.scheduler.extensions["memory_sampler"] = self
        self.scheduler.handlers["memory_sampler_start"] = self.start
        self.scheduler.handlers["memory_sampler_stop"] = self.stop
        self.samples = {}

    def start(self, client: str, measure: str, interval: float) -> str:
        """Start periodically sampling memory"""
        assert not measure.startswith("_")
        assert isinstance(getattr(self.scheduler.memory, measure), int)

        key = str(uuid.uuid4())
        self.samples[key] = []

        def sample():
            if client in self.scheduler.clients:
                ts = datetime.now().timestamp()
                nbytes = getattr(self.scheduler.memory, measure)
                self.samples[key].append((ts, nbytes))
            else:
                self.stop(key)

        pc = PeriodicCallback(sample, interval * 1000)
        self.scheduler.periodic_callbacks["MemorySampler-" + key] = pc
        pc.start()

        # Immediately collect the first sample; this also ensures there's always at
        # least one sample
        sample()

        return key

    def stop(self, key: str) -> list[tuple[float, int]]:
        """Stop sampling and return the samples"""
        pc = self.scheduler.periodic_callbacks.pop("MemorySampler-" + key)
        pc.stop()
        return self.samples.pop(key)



from __future__ import annotations

import math
from collections import defaultdict
from itertools import compress, product
from typing import TYPE_CHECKING, NamedTuple

import dask
from dask.base import tokenize
from dask.highlevelgraph import HighLevelGraph, MaterializedLayer

from distributed.exceptions import Reschedule
from distributed.shuffle._shuffle import (
    ShuffleId,
    ShuffleType,
    _get_worker_extension,
    barrier_key,
    shuffle_barrier,
)

if TYPE_CHECKING:
    import numpy as np
    from typing_extensions import TypeAlias

    import dask.array as da


ChunkedAxis: TypeAlias = tuple[float, ...]  # chunks must either be an int or NaN
ChunkedAxes: TypeAlias = tuple[ChunkedAxis, ...]
NIndex: TypeAlias = tuple[int, ...]
NSlice: TypeAlias = tuple[slice, ...]


def rechunk_transfer(
    input: np.ndarray,
    id: ShuffleId,
    input_chunk: NIndex,
    new: ChunkedAxes,
    old: ChunkedAxes,
) -> int:
    try:
        return _get_worker_extension().add_partition(
            input,
            input_partition=input_chunk,
            shuffle_id=id,
            type=ShuffleType.ARRAY_RECHUNK,
            new=new,
            old=old,
        )
    except Exception as e:
        raise RuntimeError(f"rechunk_transfer failed during shuffle {id}") from e


def rechunk_unpack(
    id: ShuffleId, output_chunk: NIndex, barrier_run_id: int
) -> np.ndarray:
    try:
        return _get_worker_extension().get_output_partition(
            id, barrier_run_id, output_chunk
        )
    except Reschedule as e:
        raise e
    except Exception as e:
        raise RuntimeError(f"rechunk_unpack failed during shuffle {id}") from e


def rechunk_p2p(x: da.Array, chunks: ChunkedAxes) -> da.Array:
    import numpy as np

    import dask.array as da

    if x.size == 0:
        # Special case for empty array, as the algorithm below does not behave correctly
        return da.empty(x.shape, chunks=chunks, dtype=x.dtype)

    old_chunks = x.chunks
    new_chunks = chunks

    def is_unknown(dim: ChunkedAxis) -> bool:
        return any(math.isnan(chunk) for chunk in dim)

    old_is_unknown = [is_unknown(dim) for dim in old_chunks]
    new_is_unknown = [is_unknown(dim) for dim in new_chunks]

    if old_is_unknown != new_is_unknown or any(
        new != old for new, old in compress(zip(old_chunks, new_chunks), old_is_unknown)
    ):
        raise ValueError(
            "Chunks must be unchanging along dimensions with missing values.\n\n"
            "A possible solution:\n  x.compute_chunk_sizes()"
        )

    old_known = [dim for dim, unknown in zip(old_chunks, old_is_unknown) if not unknown]
    new_known = [dim for dim, unknown in zip(new_chunks, new_is_unknown) if not unknown]

    old_sizes = [sum(o) for o in old_known]
    new_sizes = [sum(n) for n in new_known]

    if old_sizes != new_sizes:
        raise ValueError(
            f"Cannot change dimensions from {old_sizes!r} to {new_sizes!r}"
        )

    dsk: dict = {}
    token = tokenize(x, chunks)
    _barrier_key = barrier_key(ShuffleId(token))
    name = f"rechunk-transfer-{token}"
    transfer_keys = []
    for index in np.ndindex(tuple(len(dim) for dim in x.chunks)):
        transfer_keys.append((name,) + index)
        dsk[(name,) + index] = (
            rechunk_transfer,
            (x.name,) + index,
            token,
            index,
            chunks,
            x.chunks,
        )

    dsk[_barrier_key] = (shuffle_barrier, token, transfer_keys)

    name = f"rechunk-p2p-{token}"

    for index in np.ndindex(tuple(len(dim) for dim in chunks)):
        dsk[(name,) + index] = (rechunk_unpack, token, index, _barrier_key)

    with dask.annotate(shuffle=lambda key: key[1:]):
        layer = MaterializedLayer(dsk)
        graph = HighLevelGraph.from_collections(name, layer, dependencies=[x])

        return da.Array(graph, name, chunks, meta=x)


class ShardID(NamedTuple):
    """Unique identifier of an individual shard within an array rechunk

    When rechunking a 1d-array with two chunks into a 1d-array with a single chunk
    >>> old = ((2, 2),)  # doctest: +SKIP
    >>> new = ((4),)  # doctest: +SKIP
    >>> rechunk_slicing(old, new)  # doctest: +SKIP
    {
        # The first chunk of the old array belongs to the first
        # chunk of the new array at the first sub-index
        (0,): [(ShardID((0,), (0,)), (slice(0, 2, None),))],

        # The second chunk of the old array belongs to the first
        # chunk of the new array at the second sub-index
        (1,): [(ShardID((0,), (1,)), (slice(0, 2, None),))],
    }
    """

    #: Index of the new output chunk to which the shard belongs
    chunk_index: NIndex
    #: Index of the shard within the multi-dimensional array of shards that will be
    # concatenated into the new chunk
    shard_index: NIndex


def rechunk_slicing(
    old: ChunkedAxes, new: ChunkedAxes
) -> dict[NIndex, list[tuple[ShardID, NSlice]]]:
    """Calculate how to slice the old chunks to create the new chunks

    Returns
    -------
        Mapping of each old chunk to a list of tuples defining where each slice
        of the old chunk belongs. Each tuple consists of the index
        of the new chunk, the index of the slice within the composition of slices
        creating the new chunk, and the slice to be applied to the old chunk.
    """
    from dask.array.rechunk import intersect_chunks

    ndim = len(old)
    intersections = intersect_chunks(old, new)
    new_indices = product(*(range(len(c)) for c in new))

    slicing = defaultdict(list)

    for new_index, new_chunk in zip(new_indices, intersections):
        sub_shape = [len({slice[dim][0] for slice in new_chunk}) for dim in range(ndim)]

        shard_indices = product(*(range(dim) for dim in sub_shape))

        for shard_index, sliced_chunk in zip(shard_indices, new_chunk):
            old_index, nslice = zip(*sliced_chunk)
            slicing[old_index].append((ShardID(new_index, shard_index), nslice))
    return slicing



"""
Diagnostics for GPU memory managed by RMM (RAPIDS memory manager).
"""

from __future__ import annotations

try:
    import rmm
except ImportError:
    rmm = None


def _get_pool_size(mr):
    # if the memory resource or any of its upstreams
    # is a `PoolMemoryResource`, get its pool size
    if not isinstance(mr, rmm.mr.PoolMemoryResource):
        if hasattr(mr, "upstream_mr"):
            return _get_pool_size(mr.upstream_mr)
        else:
            return 0
    else:
        pool_size = mr.pool_size()
        return pool_size


def _get_allocated_bytes(mr):
    if not hasattr(mr, "get_allocated_bytes"):
        if hasattr(mr, "upstream_mr"):
            return _get_allocated_bytes(mr.upstream_mr)
        else:
            return 0
    else:
        return mr.get_allocated_bytes()


def real_time():
    if rmm is None:
        return {"rmm-used": None, "rmm-total": None}
    mr = rmm.mr.get_current_device_resource()
    rmm_pool_size = _get_pool_size(mr)
    rmm_used = _get_allocated_bytes(mr)
    rmm_total = max(rmm_pool_size, rmm_used)
    return {"rmm-used": rmm_used, "rmm-total": rmm_total}



from __future__ import annotations

import math
from textwrap import dedent

from bokeh.core.properties import without_property_validation
from bokeh.models import (
    BasicTicker,
    ColumnDataSource,
    HoverTool,
    NumeralTickFormatter,
    OpenURL,
    TapTool,
)
from bokeh.plotting import figure
from tornado import escape

from dask.utils import format_bytes

from distributed.dashboard.components import DashboardComponent, add_periodic_callback
from distributed.dashboard.components.scheduler import BOKEH_THEME, TICKS_1024
from distributed.dashboard.utils import update
from distributed.utils import log_errors


class RMMMemoryUsage(DashboardComponent):
    """
    GPU memory usage plot that includes information about memory
    managed by RMM. If an RMM pool is being used, shows the amount of
    pool memory utilized.
    """

    def __init__(self, scheduler, width=600, **kwargs):
        with log_errors():
            self.last = 0
            self.scheduler = scheduler
            self.source = ColumnDataSource(
                {
                    "rmm-used": [1, 2],
                    "rmm-used-half": [0.5, 1],
                    "rmm-total": [2, 4],
                    "rmm-total-half": [1, 2],
                    "external-used": [2, 1],
                    "external-used-x": [3, 4.5],
                    "worker": ["a", "b"],
                    "gpu-index": [0, 0],
                    "y": [1, 2],
                    "escaped_worker": ["a", "b"],
                    "rmm_memory_text": [
                        "RMM memory used: 1B/1B\nTotal GPU memory used: 1B/2B",
                        "RMM memory used: 1B/1B\nTotal GPU memory used: 1B/2B",
                    ],
                }
            )

            memory = figure(
                title="RMM Memory",
                tools="",
                width=int(width / 2),
                name="rmm_memory_histogram",
                **kwargs,
            )

            rect = memory.rect(
                source=self.source,
                x="rmm-used-half",
                y="y",
                width="rmm-used",
                height=1,
                color="#76B900",
                alpha=1.0,
            )
            rect.nonselection_glyph = None

            rect = memory.rect(
                source=self.source,
                x="rmm-total-half",
                y="y",
                width="rmm-total",
                height=1,
                color="#76B900",
                alpha=0.75,
            )
            rect.nonselection_glyph = None

            rect = memory.rect(
                source=self.source,
                x="external-used-x",
                y="y",
                width="external-used",
                height=1,
                color="#76B900",
                alpha=0.5,
            )
            rect.nonselection_glyph = None

            memory.axis[0].ticker = BasicTicker(**TICKS_1024)
            memory.xaxis[0].formatter = NumeralTickFormatter(format="0.0 b")
            memory.xaxis.major_label_orientation = -math.pi / 12
            memory.x_range.start = 0

            for fig in [memory]:
                fig.xaxis.minor_tick_line_alpha = 0
                fig.yaxis.visible = False
                fig.ygrid.visible = False

                tap = TapTool(
                    callback=OpenURL(url="./info/worker/@escaped_worker.html")
                )
                fig.add_tools(tap)

                fig.toolbar_location = None
                fig.yaxis.visible = False

            hover = HoverTool()
            hover.tooltips = "@worker : @rmm_memory_text"
            hover.point_policy = "follow_mouse"
            memory.add_tools(hover)

            self.memory_figure = memory

    @without_property_validation
    def update(self):
        with log_errors():
            workers = list(self.scheduler.workers.values())
            rmm_total = []
            rmm_used = []
            external_used = []
            gpu_index = []
            y = []
            worker = []
            external_used_x = []
            memory_max = 0
            gpu_total = []
            rmm_memory_text = []

            for idx, ws in enumerate(workers):
                try:
                    rmm_metrics = ws.metrics["rmm"]
                    gpu_metrics = ws.metrics["gpu"]
                    gpu_info = ws.extra["gpu"]
                except KeyError:
                    continue
                rmm_total_worker = rmm_metrics["rmm-total"]  # RMM memory only
                rmm_used_worker = rmm_metrics["rmm-used"]
                gpu_total_worker = gpu_info["memory-total"]  # All GPU memory
                gpu_used_worker = gpu_metrics["memory-used"]

                external_used_worker = gpu_used_worker - rmm_total_worker

                rmm_total.append(rmm_total_worker)
                rmm_used.append(rmm_used_worker)
                gpu_total.append(gpu_total_worker)
                external_used.append(external_used_worker)
                external_used_x.append(rmm_total_worker + external_used_worker / 2)
                worker.append(ws.address)
                gpu_index.append(idx)
                y.append(idx)

                memory_max = max(memory_max, gpu_total_worker)

                rmm_memory_text.append(
                    "RMM memory used: {}/{}\nTotal GPU memory used: {}/{}".format(
                        format_bytes(rmm_used_worker),
                        format_bytes(rmm_total_worker),
                        format_bytes(gpu_used_worker),
                        format_bytes(gpu_total_worker),
                    )
                )

            self.memory_figure.title.text = dedent(
                """\
                RMM Utilization: {} / {}
                GPU Memory: {} / {}
                """.format(
                    format_bytes(sum(rmm_used)),
                    format_bytes(sum(rmm_total)),
                    format_bytes(sum([*rmm_total, *external_used])),
                    format_bytes(sum(gpu_total)),
                )
            )

            result = {
                "rmm-total": rmm_total,
                "rmm-used": rmm_used,
                "external-used": external_used,
                "rmm-total-half": [m // 2 for m in rmm_total],
                "rmm-used-half": [m // 2 for m in rmm_used],
                "external-used-x": external_used_x,
                "worker": worker,
                "gpu-index": gpu_index,
                "y": y,
                "escaped_worker": [escape.url_escape(w) for w in worker],
                "rmm_memory_text": rmm_memory_text,
            }

            self.memory_figure.x_range.end = memory_max

            update(self.source, result)


def rmm_memory_doc(scheduler, extra, doc):
    with log_errors():
        rmm_load = RMMMemoryUsage(scheduler, sizing_mode="stretch_both")
        rmm_load.update()
        add_periodic_callback(doc, rmm_load, 100)
        doc.add_root(rmm_load.memory_figure)
        doc.theme = BOKEH_THEME



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING

from packaging.version import parse

if TYPE_CHECKING:
    import pandas as pd
    import pyarrow as pa


def check_dtype_support(meta_input: pd.DataFrame) -> None:
    import pandas as pd

    for name in meta_input:
        column = meta_input[name]
        # FIXME: PyArrow does not support complex numbers: https://issues.apache.org/jira/browse/ARROW-638
        if pd.api.types.is_complex_dtype(column):
            raise TypeError(
                f"p2p does not support data of type '{column.dtype}' found in column '{name}'."
            )
        # FIXME: PyArrow does not support sparse data: https://issues.apache.org/jira/browse/ARROW-8679
        if isinstance(column.dtype, pd.SparseDtype):
            raise TypeError("p2p does not support sparse data found in column '{name}'")


def check_minimal_arrow_version() -> None:
    """Verify that the the correct version of pyarrow is installed to support
    the P2P extension.

    Raises a RuntimeError in case pyarrow is not installed or installed version
    is not recent enough.
    """
    # First version to introduce Table.sort_by
    minversion = "7.0.0"
    try:
        import pyarrow as pa
    except ImportError:
        raise RuntimeError(f"P2P shuffling requires pyarrow>={minversion}")

    if parse(pa.__version__) < parse(minversion):
        raise RuntimeError(
            f"P2P shuffling requires pyarrow>={minversion} but only found {pa.__version__}"
        )


def convert_partition(data: bytes) -> pa.Table:
    import pyarrow as pa

    file = BytesIO(data)
    end = len(data)
    shards = []
    while file.tell() < end:
        sr = pa.RecordBatchStreamReader(file)
        shards.append(sr.read_all())
    return pa.concat_tables(shards)


def list_of_buffers_to_table(data: list[bytes]) -> pa.Table:
    """Convert a list of arrow buffers and a schema to an Arrow Table"""
    import pyarrow as pa

    return pa.concat_tables(deserialize_table(buffer) for buffer in data)


def deserialize_schema(data: bytes) -> pa.Schema:
    """Deserialize an arrow schema

    Examples
    --------
    >>> b = schema.serialize()  # doctest: +SKIP
    >>> deserialize_schema(b)  # doctest: +SKIP

    See also
    --------
    pa.Schema.serialize
    """
    import io

    import pyarrow as pa

    bio = io.BytesIO()
    bio.write(data)
    bio.seek(0)
    sr = pa.RecordBatchStreamReader(bio)
    table = sr.read_all()
    bio.close()
    return table.schema


def serialize_table(table: pa.Table) -> bytes:
    import io

    import pyarrow as pa

    stream = io.BytesIO()
    with pa.ipc.new_stream(stream, table.schema) as writer:
        writer.write_table(table)
    return stream.getvalue()


def deserialize_table(buffer: bytes) -> pa.Table:
    import pyarrow as pa

    with pa.ipc.open_stream(pa.py_buffer(buffer)) as reader:
        return reader.read_all()
